//===================================CISCKER====================================
//==================================Data Path===================================
//------------------------------------------------------------------------------
//File Name:	datapath.v
//Purpose:		Data Path, top module that instantiates all processor units
//Parameters:	None
//------------------------------------------------------------------------------
//=====================Alfredo Gualdron - Jose P. Pinilla=======================
//=================================Semillero ADT================================
//================================UPB Bucaramanga===============================
//Website:	http://semilleroadt.upbbga.edu.co/xiscker
//E-mail:	semilleroadt@gmail.com
//------------------------------------------------------------------------------

module datapath(
	input clk_0,
	input clk_90,
	input clk_180,
	input clk_270,
	input load,
	input [7:0]input_port1,
	input [7:0]input_port2,
	output [7:0] output_port1,
	output [7:0] output_port2
	);


parameter initial_address = 16'h0100;

/////////////////////////////////////////////////////////////////////////////
//=============================INSTRUCTION DECODER===========================
/////////////////////////////////////////////////////////////////////////////
reg [7:0]opcode=0;

wire [5:0] CCR;			//Condition Code Register
wire [5:0] alu_op;
wire [2:0] alu_src_a;
wire [1:0] alu_src_b;
wire [1:0] adder_src_a;
wire [1:0] adder_src_b;
wire [7:0] control; //opcode_en,pass_adder,pc_en,adder_cin,h_reset,mah_en,we,mdr_en

wire opcode_en,pass_adder,pc_en,adder_cin,h_reset,mah_en,we,mdr_en;

instruction_decoder inst_decoder (
    //in
	.opcode(opcode), 
    .load(load), 
    .clk_0(clk_0),
    .clk_90(clk_90),
	.CCR(CCR), 
	 //out
    .alu_op(alu_op), 
    .alu_operands({alu_src_a,alu_src_b}), 
    .adder_operands({adder_src_a,adder_src_b}), 
    .control({opcode_en,pass_adder,pc_en,adder_cin,h_reset,mah_en,we,mdr_en}),
    .wd_src(wd_src),
	.branch(branch),
	 //
	 .adder_alu(adder_alu), 
    .acca_src(acca_src), 
    .ccr_en(ccr_en), 
    .sp_en(sp_en), 
    .ix_en(ix_en), 
    .acca_en(acca_en), 
    .accb_en(accb_en), 
    .hidden_en(hidden_en),
	.final(final)
    );

/////////////////////////////////////////////////////////////////////////////
//===================================ADDER===================================
/////////////////////////////////////////////////////////////////////////////

wire [15:0]alu_result;
wire [15:0]adder;

//////////////////////////////////////////
reg [15:0] SP=0; //Stack Pointer

always @ (posedge clk_270)
	if(sp_en)
		if(adder_alu)
			SP <= alu_result;
		else
			SP <= adder;

//////////////////////////////////////////			
reg [15:0] IX=0; //Index Register

always @ (posedge clk_270)
	if(ix_en)
		if(adder_alu)
			IX <= alu_result;
		else
			IX <= adder;

//////////////////////////////////////////			
reg [15:0] MAH=0; //Memory Address Holder
wire[15:0] MM;

always @ (posedge clk_270)
	if(mah_en)
		MAH <= MM;

//////////////////////////////////////////		
reg [15:0] PC = initial_address; //Program Counter
wire interruption;
wire [15:0]int_address;
always@(posedge clk_270)
	if (pc_en)
		begin
			if (load)
				PC <= initial_address;
			else if (interruption)
				PC <= int_address;
			else
				PC <= adder;
		end

//////////////////////////////////////////
wire [15:0] IAB;

next_address address(
			//control
			.adder_cin(adder_cin), 
			.adder_src_a(adder_src_a),
			.adder_src_b(adder_src_b),
			.pass_adder(pass_adder),
			//in
			.PC(PC),
			.SP(SP),
			.IX(IX),
			.MAH(MAH),
			//out
			.adder(adder),
			.IAB(IAB) //Internal Address Bus
			);
			
//////////////////////////////////////////		
reg [15:0] MAR=0;	//Memory Address Register

always @ (posedge clk_180)
	MAR <= IAB;
			
/////////////////////////////////////////////////////////////////////////////
//===================================MEMORY==================================
/////////////////////////////////////////////////////////////////////////////

reg [15:0] MDR=0;

always @ (posedge clk_270)
	if (mdr_en)
		MDR <= alu_result;
		
wire select_MMreg_or_memory = MAR[15];
wire we_memory	= we & ~select_MMreg_or_memory;
wire we_MMreg	= we & select_MMreg_or_memory;


wire [7:0] Mem_data;

memory mem(
		//control
		.we(we_memory),
		.clk_90(clk_90),
		.wd_src(wd_src),
		//in
		.MDR(MDR),
		.MAR(MAR[14:0]),
		//out
		.Mem_data(Mem_data)
		);
		
//=======Memory Mapped Registers=======
wire [7:0] Reg_data;
wire [7:0]int_ctrl;
wire int_timer1;
wire int_timer2;

MM_registers MM_registers_inst (
		//Inputs
		.clk(clk_90) ,				// Same as Memory Clock
		.tmr_clk(clk_0) ,			//Timers counter clock
		.we(we_MMreg) ,				// Write Enable
		.data(MDR[7:0]) ,			// Input Data to be written
		.address(MAR[7:0]) ,		// Registers Address
		.ip1(input_port1) ,			// Input Port 1
		.ip2(input_port2) ,			// Input Port 2
		//Outputs
		.op1(output_port1) ,
		.op2(output_port2) ,
		.int_ctrl(int_ctrl) ,		// Interruption Control Register
		.int_timer1(int_timer1) ,
		.int_timer2(int_timer2) ,
		.out(Reg_data) 				// Read Data from Register
		);

reg [7:0]Reg_data_270=0;
reg select_MMreg_or_memory_270;
always @ (posedge clk_270)
	begin
		Reg_data_270<=Reg_data;
		select_MMreg_or_memory_270<=select_MMreg_or_memory;
	end

wire [7:0] ML = (select_MMreg_or_memory_270) ? Reg_data_270 : Mem_data;

//========Memory extended=========
wire sign_ext = branch & MM[7]; //Relative data on Branch is sign extended

wire h_clear = h_reset & ~sign_ext;
reg [7:0] MH = 0;		

always @ (posedge clk_90, posedge h_clear, posedge sign_ext)
	if (h_clear)
		MH <= 8'b0;
	else
		if (sign_ext)
			MH <= 8'hFF;
		else
			MH <= ML;
		
assign MM = {MH,ML};

/////////////////////////////////////////////////////////////////////////////
//===================================OPCODE==================================
/////////////////////////////////////////////////////////////////////////////

always @ (posedge clk_180)
	if (opcode_en)
		if (interruption)
			opcode <= 8'h2F;	//SWI: Forced Software Interruption
		else
			opcode <= MM[7:0];

reg [7:0] opcode_clk0;

always @ (posedge clk_0)
	if (final)
		opcode_clk0 <= opcode;
		
/////////////////////////////////////////////////////////////////////////////
//===================================ALU=====================================
/////////////////////////////////////////////////////////////////////////////	

wire wd_ccr = (opcode_clk0==8'h2F); //During Interruption

reg [7:0] ACCA=0;
reg [7:0] ACCB=0;
reg [7:0] hidden=0;
wire[15:0] ACCD = {ACCA,ACCB};
wire [2:0] ccr_data_a;
wire [1:0] ccr_data_b;

wire [15:0] ACCA_CCR = wd_ccr ?  {2'b0,CCR,ACCA} : {8'b0,ACCA};

ALU Arithmetic_Logic_Unit(
			//control
			.alu_src_a(alu_src_a), 	//3
			.alu_src_b(alu_src_b), 	//2
			.alu_op(alu_op), 		//6
			//in
			.ACCA(ACCA_CCR),
			.ACCB(ACCB),
			.ACCD(ACCD),
			.MM(MM),
			.PC(PC),
			.SP(SP),
			.IX(IX),	
			.hidden(hidden),
			//out
			.alu_result(alu_result),
			.ccr_data_a(ccr_data_a),
			.ccr_data_b(ccr_data_b)
			);
///////MULTIPLIER AND DIVIDER////////
wire [7:0] acca_result, accb_result; // it will have the result in the 8th clock cycle
wire mult_div_sel = (opcode_clk0 == 8'h09) ? 1'b1 : 1'b0 ; //  Mul=1 - Div=0	(select the operation)
wire mult_div_opr = (opcode_clk0 == 8'h09 | opcode_clk0 == 8'h08); // is high when mult or div operation is resolving

multiplier_divider multiplier_divider_inst (
	.acca(alu_result),						//input ACCA or ZERO
	.accb(ACCB),                    		//input 
	.hidden(hidden),                  		//input 
	.mult_div(mult_div_sel),                //input
	.acca_result(acca_result),            	//output
	.accb_result(accb_result)           	//output
	);

defparam multiplier_divider_inst.width = 8;

////////////////////////////////////////
always @ (posedge clk_270)
	if(acca_en)
		if (mult_div_opr)
			ACCA <= acca_result;
		else if (acca_src)
			ACCA <= alu_result[7:0];
		else
			ACCA <= alu_result[15:8];
/////////////////////////////////////////
always @ (posedge clk_270)
	if(hidden_en)
		hidden <= alu_result[7:0];
////////////////////////////////////////
always @ (posedge clk_270)
	if(accb_en)
		if (mult_div_opr)
			ACCB <= accb_result;
		else
			ACCB <= alu_result[7:0];

/////////////////////////////////////////////////////////////////////////////
//===================================CCR=====================================
/////////////////////////////////////////////////////////////////////////////	

CCR condition_code_register (
		//control
		.ccr_en(ccr_en),
		.clk(clk_270),
		//in
		.r(alu_result),
		.a(ccr_data_a),
		.b(ccr_data_b),
		.opcode(opcode_clk0),
		//out
		.CCR(CCR)
		);

/////////////////////////////////////////////////////////////////////////////
//===========================INTERRUPTION CONTROL============================
/////////////////////////////////////////////////////////////////////////////
wire rti = (opcode== 8'h2D);
wire swi = (opcode== 8'h2F);

interrupt Interruption_Controller (
		//Inputs
		.swi(swi),
		.rti(rti),
		.I_mask(CCR[3]),
		.int_ext1(input_port1[7]),
		.int_ext2(input_port1[6]),
		.int_timer1(int_timer1),
		.int_timer2(int_timer2),
		.int_ctrl(int_ctrl),
		.clk(clk_0),
		//Outputs
		.interruption(interruption),
		.int_address(int_address)
		);
		
endmodule
